#include "motor_thrd.h"
#include "motor_thrd_api.h"
#include "tof_thrd_api.h"
#include "sol_thrd_api.h"

#define MOTOR_MODEOFOPERATION_MANUAL 1
#define MOTOR_MODEOFOPERATION_REACHDEST 2

static gpt_instance_ctrl_t* motorGpt[2] = { &g_motor1_ctrl, &g_motor2_ctrl };
static const timer_cfg_t* motorGptConfig[2] = { &g_motor1_cfg, &g_motor2_cfg };
static int modeOfOperation = MOTOR_MODEOFOPERATION_MANUAL;

static uint16_t targetTof1 = 150;
static uint16_t targetTof2 = 150;
static int direction1 = MOTOR_DIRECTION_FORWARD;
static int direction2 = MOTOR_DIRECTION_BACKWARD;
static TickType_t lastReachProgressPrint = 0;

static void Motor_Init(int motorNumber) {
	fsp_err_t fStatus;

	fStatus = R_GPT_Open(motorGpt[motorNumber], motorGptConfig[motorNumber]);
	if (fStatus) {
		__BKPT(1);
	}

	Motor_TurnOff(motorNumber);

	fStatus = R_GPT_Start(motorGpt[motorNumber]);
	if (fStatus) {
		__BKPT(1);
	}
}

static void Motor_TurnOffInternal(int motorNumber) {
	R_GPT_OutputDisable(motorGpt[motorNumber], GPT_IO_PIN_GTIOCA);
	R_GPT_OutputDisable(motorGpt[motorNumber], GPT_IO_PIN_GTIOCB);
}

int Motor_TurnOff(int motorNumber) {
	if (modeOfOperation != MOTOR_MODEOFOPERATION_MANUAL) {
		return 1;
	}

	Motor_TurnOffInternal(motorNumber);
	return 0;
}

static void Motor_SetInternal(int motorNumber, int direction, int speed) {
	fsp_err_t fStatus;

	if (direction == MOTOR_DIRECTION_FORWARD) {
		fStatus = R_GPT_DutyCycleSet(motorGpt[motorNumber], (uint32_t)((100 - speed) * 100), GPT_IO_PIN_GTIOCA);
		if (fStatus) {
			__BKPT(1);
		}

		fStatus = R_GPT_OutputEnable(motorGpt[motorNumber], GPT_IO_PIN_GTIOCA);
		if (fStatus) {
			__BKPT(1);
		}

		fStatus = R_GPT_OutputDisable(motorGpt[motorNumber], GPT_IO_PIN_GTIOCB);
		if (fStatus) {
			__BKPT(1);
		}
	} else {
		fStatus = R_GPT_DutyCycleSet(motorGpt[motorNumber], (uint32_t)((100 - speed) * 100), GPT_IO_PIN_GTIOCB);
		if (fStatus) {
			__BKPT(1);
		}

		fStatus = R_GPT_OutputEnable(motorGpt[motorNumber], GPT_IO_PIN_GTIOCB);
		if (fStatus) {
			__BKPT(1);
		}

		fStatus = R_GPT_OutputDisable(motorGpt[motorNumber], GPT_IO_PIN_GTIOCA);
		if (fStatus) {
			__BKPT(1);
		}
	}
}

int Motor_Set(int motorNumber, int direction, int speed) {
	if (modeOfOperation != MOTOR_MODEOFOPERATION_MANUAL) {
		return 1;
	}

	Motor_SetInternal(motorNumber, direction, speed);
	return 0;
}

int MotorSet_ReachDestination_Start(uint16_t tof1Pos, uint16_t tof2Pos) {
	if (modeOfOperation != MOTOR_MODEOFOPERATION_MANUAL) {
		return 1;
	}

	targetTof1 = tof1Pos;
	targetTof2 = tof2Pos;
	modeOfOperation = MOTOR_MODEOFOPERATION_REACHDEST;

	Motor_TurnOffInternal(0);
	Motor_TurnOffInternal(1);

	uint16_t actualTof1, actualTof2;
	TickType_t timestamp1, timestamp2;

	ToFManager_GetDistance(0, &timestamp1, &actualTof1);
	ToFManager_GetDistance(1, &timestamp2, &actualTof2);

	TickType_t now = xTaskGetTickCount();

	if (now - timestamp1 > 1000 || now - timestamp2 > 100) {
		return 2;
	}

	if (targetTof1 > actualTof1) {
		direction1 = MOTOR_DIRECTION_FORWARD;
	} else {
		direction1 = MOTOR_DIRECTION_BACKWARD;
	}

	if (targetTof2 > actualTof2) {
		direction2 = MOTOR_DIRECTION_FORWARD;
	} else {
		direction2 = MOTOR_DIRECTION_BACKWARD;
	}

	lastReachProgressPrint = now;
	SEGGER_RTT_printf(0, "=============== REACH START ===============\r\n");
	SEGGER_RTT_printf(0, "TOF1: %d -> %d (direction: %s)\r\n", actualTof1, targetTof1, direction1 == MOTOR_DIRECTION_FORWARD ? "FWD" : "BWD");
	SEGGER_RTT_printf(0, "TOF2: %d -> %d (direction: %s)\r\n", actualTof2, targetTof2, direction2 == MOTOR_DIRECTION_FORWARD ? "FWD" : "BWD");
	SEGGER_RTT_printf(0, "===========================================\r\n");

	return 0;
}

static void MotorSet_ReachDestination_Process() {
	uint16_t actualTof1, actualTof2;
	TickType_t timestamp1, timestamp2;

	ToFManager_GetDistance(0, &timestamp1, &actualTof1);
	ToFManager_GetDistance(1, &timestamp2, &actualTof2);

	TickType_t now = xTaskGetTickCount();

	if (now - timestamp1 > 1000 || now - timestamp2 > 100) {
		SEGGER_RTT_printf(0, "ToF failure while reaching destination. Waiting and retrying.\r\n");

		/*
		modeOfOperation = MOTOR_MODEOFOPERATION_MANUAL;
		*/

		Motor_TurnOffInternal(1);
		Motor_TurnOffInternal(0);

		vTaskDelay(20);

		return;
	}

	int iActualTof1 = (int)actualTof1;
	int iActualTof2 = (int)actualTof2;
	int iTargetTof1 = (int)targetTof1;
	int iTargetTof2 = (int)targetTof2;

	int remaining1, remaining2;
	int isMovingAny = 0;

	if (direction1 == MOTOR_DIRECTION_FORWARD) {
		remaining1 = iTargetTof1 - iActualTof1;
	} else {
		remaining1 = iActualTof1 - iTargetTof1;
	}

	if (direction2 == MOTOR_DIRECTION_FORWARD) {
		remaining2 = iTargetTof2 - iActualTof2;
	} else {
		remaining2 = iActualTof2 - iTargetTof2;
	}

	if (remaining1 > 0) {
		Motor_SetInternal(0, direction1, 100);
		isMovingAny = 1;
	} else {
		Motor_TurnOffInternal(0);
	}

	if (remaining2 > 30) {
		Motor_SetInternal(1, direction2, 100);
		isMovingAny = 1;
	} else if (remaining2 > 0) {
		Motor_SetInternal(1, direction2, 75);
		isMovingAny = 1;
	} else {
		Motor_TurnOffInternal(1);
	}

	if (isMovingAny == 0) {
		modeOfOperation = MOTOR_MODEOFOPERATION_MANUAL;
		Solenoid_Trigger();
		SEGGER_RTT_printf(0, "=============== REACH SUCCESS ===============\r\n");
		return;
	}

	if (now - lastReachProgressPrint > 3000) {
		lastReachProgressPrint = now;
		SEGGER_RTT_printf(0, "=============== REACH PROGRESS ===============\r\n");
		SEGGER_RTT_printf(0, "TOF1: %d -> %d (direction: %s)\r\n", actualTof1, targetTof1, direction1 == MOTOR_DIRECTION_FORWARD ? "FWD" : "BWD");
		SEGGER_RTT_printf(0, "TOF2: %d -> %d (direction: %s)\r\n", actualTof2, targetTof2, direction2 == MOTOR_DIRECTION_FORWARD ? "FWD" : "BWD");
		SEGGER_RTT_printf(0, "==============================================\r\n");
	}
}

int MotorSet_ReachDestination_GetStatus() {
	return 0;
}

void motor_thrd_entry(void* pvParameters) {
	FSP_PARAMETER_NOT_USED(pvParameters);

	Motor_Init(0);
	Motor_Init(1);

	while (1) {
		if (modeOfOperation == MOTOR_MODEOFOPERATION_REACHDEST) {
			MotorSet_ReachDestination_Process();
			vTaskDelay(1);
		} else {
			vTaskDelay(10);
		}
	}
}

