#include "common_utils.h"
#include "net_thrd.h"
#include "FreeRTOS_IP.h"
#include "FreeRTOS_IP_Private.h"
#include "FreeRTOS_DHCP.h"
#include "FreeRTOS_Sockets.h"
#include <string.h>
#include "tof_thrd_api.h"
#include "motor_thrd_api.h"
#include "sol_thrd_api.h"

static uint8_t ucMACAddress[6] = { 0x00, 0x11, 0x22, 0x33, 0x44, 0x55 };
static uint8_t ucIPAddress[4] = { 0,0,0,0 };
static uint8_t ucNetMask[4] = { 0,0,0,0 };
static uint8_t ucGatewayAddress[4] = { 0,0,0,0 };
static uint8_t ucDNSServerAddress[4] = { 0,0,0,0 };

extern NetworkAddressingParameters_t xNetworkAddressing;

static Socket_t httpSrv;
static int isHttpServerRunning = 0;

#define ETHERNET_LINK_DOWN 1
#define ETHERNET_LINK_UP 0
#define IP_LINK_DOWN 2
#define IP_LINK_UP 0

#define SUCCESS 0

uint32_t isNetworkUp(void);

static void startHttpServer() {
	BaseType_t status = pdFALSE;

	httpSrv = FreeRTOS_socket(FREERTOS_AF_INET, FREERTOS_SOCK_STREAM, FREERTOS_IPPROTO_TCP);
	if (httpSrv == FREERTOS_INVALID_SOCKET) {
		APP_ERR_PRINT("Socket creation failed");
		APP_ERR_TRAP(status);
	}

	TickType_t recvTimeout = 1;
	FreeRTOS_setsockopt(httpSrv, 0, FREERTOS_SO_RCVTIMEO, &recvTimeout, sizeof(recvTimeout));

	struct freertos_sockaddr bindAddr;
	bindAddr.sin_len = sizeof(bindAddr);
	bindAddr.sin_addr = 0;
	bindAddr.sin_family = FREERTOS_AF_INET;
	bindAddr.sin_port = FreeRTOS_htons(80);
	status = FreeRTOS_bind(httpSrv, &bindAddr, sizeof(bindAddr));
	if (status) {
		APP_ERR_PRINT("Socket bind failed");
		APP_ERR_TRAP(status);
	}

	status = FreeRTOS_listen(httpSrv, 10);
	if (status) {
		APP_ERR_PRINT("Socket bind failed");
		APP_ERR_TRAP(status);
	}

	APP_PRINT("HTTP Server started.");
}

static void CloseClientConnection(Socket_t clientSocket) {
	TickType_t shutdownStartTime;

	FreeRTOS_shutdown(clientSocket, FREERTOS_SHUT_RDWR);
	shutdownStartTime = xTaskGetTickCount();
	do {
		uint8_t dummy = 0;
		if (FreeRTOS_recv(clientSocket, &dummy, sizeof(dummy), 0) < 0) {
			break;
		}
	} while ((xTaskGetTickCount() - shutdownStartTime) < pdMS_TO_TICKS(5000));

	FreeRTOS_closesocket(clientSocket);
	vTaskDelete(NULL);
}

static void SendString(Socket_t clientSocket, char* str) {
	size_t len = strlen(str);
	BaseType_t status = FreeRTOS_send(clientSocket, str, len, 0);
	if ((size_t)status != len) {
		SEGGER_RTT_printf(0, "FreeRTOS_send returned unexpected number of bytes. Expected: %d, Returned: %d\r\n", len, status);
		CloseClientConnection(clientSocket);
	}
}

static void SendHeader(Socket_t clientSocket, char* name, char* value) {
	SendString(clientSocket, name);
	SendString(clientSocket, ": ");
	SendString(clientSocket, value);
	SendString(clientSocket, "\r\n");
}


static void SendErrorResponse(Socket_t clientSocket, int statusCode, char* statusMessage) {
	SEGGER_RTT_printf(0, "Generating HTTP error response\r\n");
	char statusMessageStr[64];
	snprintf(statusMessageStr, sizeof(statusMessageStr), "HTTP/1.0 %d %s\r\n", statusCode, statusMessage);
	SendString(clientSocket, statusMessageStr);
	SendHeader(clientSocket, "Content-Type", "text/plain");
	SendHeader(clientSocket, "Connection", "close");
	SendString(clientSocket, "\r\n");
	SendString(clientSocket, statusMessage);
	CloseClientConnection(clientSocket);
}

static void ProcessIndexRequest(Socket_t clientSocket) {
	SEGGER_RTT_printf(0, "Processing HTTP request to index page\r\n");
	SendString(clientSocket, "HTTP/1.0 200 Ok\r\n");
	SendHeader(clientSocket, "Content-Type", "text/html");
	SendHeader(clientSocket, "Connection", "close");
	SendString(clientSocket, "\r\n");
	SendString(clientSocket, "Duratool Storage Box Upgrade - Box Opener - Firmware version 1.0");
}

static void ProcessPushRequest(Socket_t clientSocket) {
	SEGGER_RTT_printf(0, "Processing HTTP push request\r\n");

	Solenoid_Trigger();

	SendString(clientSocket, "HTTP/1.0 200 Ok\r\n");
	SendHeader(clientSocket, "Content-Type", "text/html");
	SendHeader(clientSocket, "Connection", "close");
	SendString(clientSocket, "\r\n");
	SendString(clientSocket, "OK");
}

static void ProcessToFDataRequest(Socket_t clientSocket) {
	SEGGER_RTT_printf(0, "Processing HTTP request to ToF Data page\r\n");
	SendString(clientSocket, "HTTP/1.0 200 Ok\r\n");
	SendHeader(clientSocket, "Content-Type", "application/json");
	SendHeader(clientSocket, "Connection", "close");
	SendString(clientSocket, "\r\n");

	uint16_t distance1, distance2;
	TickType_t sampleTime1, sampleTime2;
	ToFManager_GetDistance(0, &sampleTime1, &distance1);
	ToFManager_GetDistance(1, &sampleTime2, &distance2);

	int val1 = distance1;
	int val2 = distance2;

	TickType_t now = xTaskGetTickCount();

	if (now - sampleTime1 > 1000) {
		val1 = -1;
	}

	if (now - sampleTime2 > 1000) {
		val2 = -1;
	}

	char buffer[256];
	snprintf(buffer, sizeof(buffer), "\{ \"tof1\": \"%d\", \"tof2\": \"%d\" }", val1, val2);

	SendString(clientSocket, buffer);
}

static void ProcessMotorRequest(Socket_t clientSocket, int motorNumber, int direction, int speed) {
	SEGGER_RTT_printf(0, "Processing HTTP motor manual motion request\r\n");

	int iStatus;

	iStatus = Motor_Set(motorNumber, direction, speed);
	if (iStatus) {
		SEGGER_RTT_printf(0, "Motor is busy.\r\n");
		SendErrorResponse(clientSocket, 429, "Motor is busy.");
		return;
	}

	vTaskDelay(3000);
	Motor_TurnOff(motorNumber);

	SendString(clientSocket, "HTTP/1.0 200 Ok\r\n");
	SendHeader(clientSocket, "Content-Type", "text/plain");
	SendHeader(clientSocket, "Connection", "close");
	SendString(clientSocket, "\r\n");
	SendString(clientSocket, "OK");
}

static void ProcessReachRequest(Socket_t clientSocket, char* req) {
	SEGGER_RTT_printf(0, "Processing HTTP reach destination request\r\n");

	int tof2, tof1;
	int iStatus;

	iStatus = sscanf(req, "/reach/%d/%d", &tof2, &tof1);
	if (iStatus != 2) {
		SEGGER_RTT_printf(0, "Set datetime scanf failed\r\n");
		SendErrorResponse(clientSocket, 400, "Invalid parameters.");
		return;
	}

	if (tof1 < 0 || tof1 > 400) {
		SEGGER_RTT_printf(0, "Argument 2 out of allowed range.\r\n");
		SendErrorResponse(clientSocket, 400, "Argument 2 out of allowed range.");
		return;
	}

	if (tof2 < 0 || tof2 > 300) {
		SEGGER_RTT_printf(0, "Argument 1 out of allowed range.\r\n");
		SendErrorResponse(clientSocket, 400, "Argument 1 out of allowed range.");
		return;
	}

	iStatus = MotorSet_ReachDestination_Start((uint16_t)tof1, (uint16_t)tof2);
	if (iStatus == 1) {
		SEGGER_RTT_printf(0, "Motor is busy.\r\n");
		SendErrorResponse(clientSocket, 429, "Motor is busy.");
		return;
	} else if (iStatus == 2) {
		SEGGER_RTT_printf(0, "Sensor failure.\r\n");
		SendErrorResponse(clientSocket, 500, "Sensor failure.");
		return;
	}

	SendString(clientSocket, "HTTP/1.0 200 Ok\r\n");
	SendHeader(clientSocket, "Content-Type", "text/plain");
	SendHeader(clientSocket, "Connection", "close");
	SendString(clientSocket, "\r\n");
	SendString(clientSocket, "OK");
}

static void ProcessRequest(Socket_t clientSocket, char* req) {
	SEGGER_RTT_printf(0, "Processing HTTP request to %s\r\n", req);
	if (strcmp(req, "/") == 0) {
		ProcessIndexRequest(clientSocket);
	} else if (strcmp(req, "/tofdata") == 0) {
		ProcessToFDataRequest(clientSocket);
	} else if (strcmp(req, "/m1fwd") == 0) {
		ProcessMotorRequest(clientSocket, 0, MOTOR_DIRECTION_FORWARD, 100);
	} else if (strcmp(req, "/m1bwd") == 0) {
		ProcessMotorRequest(clientSocket, 0, MOTOR_DIRECTION_BACKWARD, 100);
	} else if (strcmp(req, "/m2fwd") == 0) {
		ProcessMotorRequest(clientSocket, 1, MOTOR_DIRECTION_FORWARD, 80);
	} else if (strcmp(req, "/m2bwd") == 0) {
		ProcessMotorRequest(clientSocket, 1, MOTOR_DIRECTION_BACKWARD, 80);
	} else if (strncmp(req, "/reach/", 7) == 0) {
		ProcessReachRequest(clientSocket, req);
	}  else if (strcmp(req, "/push") == 0) {
		ProcessPushRequest(clientSocket);
	} else {
		SendErrorResponse(clientSocket, 404, "Not Found");
	}
	CloseClientConnection(clientSocket);
}

static void HandleHttpClient(void* pvParameters) {
	uint8_t message[ipconfigTCP_MSS];
	Socket_t clientSocket = (Socket_t)pvParameters;
	BaseType_t receivedBytes;

	static const TickType_t timeout = pdMS_TO_TICKS(5000);

	FreeRTOS_setsockopt(clientSocket, 0, FREERTOS_SO_RCVTIMEO, &timeout, sizeof(timeout));
	FreeRTOS_setsockopt(clientSocket, 0, FREERTOS_SO_SNDTIMEO, &timeout, sizeof(timeout));

	receivedBytes = FreeRTOS_recv(clientSocket, message, sizeof(message), 0);
	if (receivedBytes >= 3) {
		if (message[0] == 'G' && message[1] == 'E' && message[2] == 'T' && message[3] == ' ') {
			char req[32];
			BaseType_t i = 0;
			while (i < (BaseType_t)sizeof(req) && receivedBytes > i + 4) {
				if (message[i + 4] == ' ') {
					req[i] = '\0';
					ProcessRequest(clientSocket, req);
				} else {
					req[i] = (char)message[i + 4];
				}
				i++;
			}
			SEGGER_RTT_printf(0, "Processing HTTP request failed because requested resource is too long\r\n", req);
			SendErrorResponse(clientSocket, 404, "Not Found");
		} else {
			SEGGER_RTT_printf(0, "Processing HTTP request because first line do not start with GET\r\n");
			SendErrorResponse(clientSocket, 405, "Method not allowed");
		}
	} else {
		CloseClientConnection(clientSocket);
	}
}

static void handleHttpClients() {
	struct freertos_sockaddr clientAddr;
	BaseType_t status = pdFALSE;
	uint32_t clientAddrSize = sizeof(clientAddr);
	Socket_t client = FreeRTOS_accept(httpSrv, &clientAddr, &clientAddrSize);
	if (client != NULL) {
		SEGGER_RTT_printf(0, "New client connected\r\n");
		status = xTaskCreate(HandleHttpClient, "HTTP Cli", 1024, (void*)client, configMAX_PRIORITIES - 1, NULL);
		if (status != pdPASS) {
			SEGGER_RTT_printf(0, "Unable to start thread handling HTTP client.\r\n");
		}
	} else {
		vTaskDelay(1);
	}
}

void net_thrd_entry(void* pvParameters) {
	BaseType_t status = pdFALSE;
	uint32_t previousNetworkStatus = 0xFFFFFFFF;

	FSP_PARAMETER_NOT_USED(pvParameters);

	status = FreeRTOS_IPInit(ucIPAddress, ucNetMask, ucGatewayAddress, ucDNSServerAddress, ucMACAddress);
	if (status != pdTRUE) {
		APP_ERR_PRINT("FreeRTOS_IPInit Failed");
		APP_ERR_TRAP(status);
	}

	SEGGER_RTT_printf(0, "IP stack was successfully initialized.\r\n");

	while (true) {
		uint32_t currentNetowrkStatus = isNetworkUp();

		if (previousNetworkStatus != currentNetowrkStatus) {
			if (currentNetowrkStatus == SUCCESS) {
				SEGGER_RTT_printf(0, "Network is UP now\r\n");
			} else {
				SEGGER_RTT_printf(0, "Network is DOWN now\r\n");
			}
		}

		previousNetworkStatus = currentNetowrkStatus;

		if (SUCCESS == currentNetowrkStatus) {
			if (!isHttpServerRunning) {
				isHttpServerRunning = 1;
				startHttpServer();
			} else {
				handleHttpClients();
			}
		} else {
		}

		vTaskDelay(1);
	}
}

eDHCPCallbackAnswer_t xApplicationDHCPHook(eDHCPCallbackPhase_t eDHCPPhase, uint32_t ulIPAddress) {
	char ipString[16];
	char netmaskString[16];
	char gwString[16];

	if (eDHCPPhase == eDHCPPhasePreRequest) {
		xNetworkAddressing.ulDefaultIPAddress = ulIPAddress;

		FreeRTOS_inet_ntoa(xNetworkAddressing.ulDefaultIPAddress, ipString);
		FreeRTOS_inet_ntoa(xNetworkAddressing.ulNetMask, netmaskString);
		FreeRTOS_inet_ntoa(xNetworkAddressing.ulGatewayAddress, gwString);
		SEGGER_RTT_printf(0, "DHCP assigned IP address: %s, netmask: %s, gateway: %s\r\n", ipString, netmaskString, gwString);
	}
	return eDHCPContinue;
}

uint32_t isNetworkUp(void) {
	fsp_err_t eth_link_status = FSP_ERR_NOT_OPEN;
	BaseType_t networkUp = pdFALSE;
	uint32_t network_status = (IP_LINK_UP | ETHERNET_LINK_UP);

	networkUp = FreeRTOS_IsNetworkUp();
	eth_link_status = R_ETHER_LinkProcess(g_ether0.p_ctrl);

	if ((FSP_SUCCESS == eth_link_status) && (pdTRUE == networkUp)) {
		return network_status;
	} else {
		if (FSP_SUCCESS != eth_link_status) {
			network_status |= ETHERNET_LINK_DOWN;
		} else if (FSP_SUCCESS == eth_link_status) {
			network_status |= ETHERNET_LINK_UP;
		}

		if (pdTRUE != networkUp) {
			network_status |= IP_LINK_DOWN;
		} else if (pdTRUE == networkUp) {
			network_status |= IP_LINK_UP;
		}
		return network_status;
	}
}

const char* pcApplicationHostnameHook(void) {
	return "Power Consumption Logger";
}

// this crazy function come from Renesas example
uint32_t ulRand() {
	uint32_t ulResult = ((((uint32_t)rand()) & 0x7fffuL)) | ((((uint32_t)rand()) & 0x7fffuL) << 15) | ((((uint32_t)rand()) & 0x0003uL) << 30);
	return ulResult;
}

// this crazy function come from Renesas example
uint32_t ulApplicationGetNextSequenceNumber(uint32_t ulSourceAddress, uint16_t usSourcePort, uint32_t ulDestinationAddress, uint16_t usDestinationPort) {
	return ((ulSourceAddress + ulDestinationAddress + usSourcePort + usDestinationPort) && ulRand());
}
